/*
 * Calculate enclosed mass of a convex polygon
 * 2015-09-29: created by Abdalla Ahmed
 * 2020-06-20: Edited for the minimum requirements of pinwheel tiling.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>

struct Point {
    double x, y;
};

typedef std::vector<Point> Points;

inline double frac(const double &x) { return x - floor(x); }

inline double crossProduct(const Point &ref, const Point &p1, const Point &p2) {
    return (p1.x - ref.x) * (p2.y - ref.y)
    - (p1.y - ref.y) * (p2.x - ref.x);
}

inline Point intersect(Point &p1, Point &p2, Point &p3, Point &p4) {            // See https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection
    double x12 = p1.x - p2.x, y12 = p1.y - p2.y;
    double x34 = p3.x - p4.x, y34 = p3.y - p4.y;
    double xy12 = p1.x * p2.y - p1.y * p2.x;
    double xy34 = p3.x * p4.y - p3.y * p4.x;
    double det = 1.0 / (x12 * y34 - y12 * x34);
    return {det * (xy12 * x34 - xy34 * x12), det * (xy12 * y34 - xy34 * y12)};
}


inline bool isConvex(Point &p1, Point &p2, Point &p3, Point &p4) {
    bool result = true;
    if (
        (crossProduct(p1, p2, p4) < 0) ||
        (crossProduct(p2, p3, p1) < 0) ||
        (crossProduct(p3, p4, p2) < 0) ||
        (crossProduct(p4, p1, p3) < 0)
    ) {
        return false;
    }
    return true;
}

double area(Points &p, bool lastEqFirst = false) {
    int n = p.size();
    double a = 0;
    for (int i = 0; i < n-1; i++) {
        a += p[i].x * p[i+1].y - p[i].y * p[i+1].x;
    }
    if (!lastEqFirst) {
        a += p[n-1].x * p[0].y - p[n-1].y * p[0].x;
    }
    return a / 2.0;
}

inline bool samePixel(Point &p1, Point&p2) {
    return (int(p1.x) == int(p2.x)) && (int(p1.y) == int(p2.y));
}

// See https://en.wikipedia.org/wiki/Sutherland%E2%80%93Hodgman_algorithm
inline bool isInsideEdge(
    const Point &p,
    const Point &edgeStart,
    const Point &edgeEnd
) {
    return ( crossProduct(p, edgeStart, edgeEnd) > 0 );
}

Points clipByEdge(Points &p, Point &edgeStart, Point &edgeEnd) {
    Points result;
    int startIndex = p.size() - 1;
    for (int endIndex = 0; endIndex < p.size(); endIndex++) {
        Point start = p[startIndex];
        Point end = p[endIndex];
        if (isInsideEdge(end, edgeStart, edgeEnd)) {
            if (!isInsideEdge(start, edgeStart, edgeEnd)) {
                result.push_back(intersect(start, end, edgeStart, edgeEnd));
            }
            result.push_back(end);
        }
        else if (isInsideEdge(start, edgeStart, edgeEnd)) {
            result.push_back(intersect(start, end, edgeStart, edgeEnd));
        }
        startIndex = endIndex;
    }
    return result;
}

Points crop(const Points &subject, const Points &boundry) {                     // See https://en.wikipedia.org/wiki/Sutherland%E2%80%93Hodgman_algorithm
    Points result = subject;
    Point edgeStart = boundry.back();
    for (int i = 0; i < boundry.size(); i++) {
        Point edgeEnd = boundry[i];
        result = clipByEdge(result, edgeStart, edgeEnd);
        edgeStart = edgeEnd;
    }
    return result;
}

double vertexPixelArea(Points &v, int i) {                                      // Returns covered ratio of the pixel at v[i]
    int xl = int(v[i].x), yb = int(v[i].y);
    Points BB = {{xl,yb}, {xl+1, yb}, {xl+1, yb+1}, {xl, yb+1}};
    Points intersection = crop(v, BB);
    return intersection.size() > 2 ? area(intersection) : 0;
}

class CDensityMap {
    int w, h, w1;
    double sum, avg, wxh_inv;
    double *bmp;                                                                // The actual bitmap, scaled to [0-1]
    double *accum;                                                              // Map of accumulated mass in columns.
    int *leftmost, *rightmost;                                                  // Define horizontal strips in polygons
    int sentinelx, sentinely;
    double excluded;                                                            // Accumulate uncovered mass
    inline void exclude(int &x, int &y, double &uncovered) {
        if (x != sentinelx || y != sentinely) {
            excluded += uncovered * getpixel(x, y);
        }
    }
public:
    CDensityMap(std::string fname, double taint);                               // 256 grades, no comments, no errors.
    double digitizeEdge(Point p1, Point p2);
    double mass(Points p);                                                      // Last point MUST be the same as first
    double mass(const Point &p1, const Point &p2, const Point &p3);
    inline double getpixel(const int &x, const int &y) {
        if (x < 0 || x >= w || y < 0 || y >= h) return 0;
        return bmp[y * w + x];
    }
    inline double getpixel(const Point &p) {
        int x = int(p.x), y = int(p.y);
        return getpixel(x, y);
    }
};

CDensityMap::CDensityMap(std::string fileName, double taint) {
    FILE *pgmfile = fopen(fileName.c_str(), "r");
    if (!pgmfile) {
        fprintf(stderr, "Failed to open file %s\n", fileName);
        exit(1);
    }
    int grades, x, y;
    fscanf(pgmfile, "P2 %d %d %d", &w, &h, &grades);
    w1 = w + 2;                                                                 // A 1 for padding and another as a margin.
    bmp = new double [h*w];
    accum = new double [h * w1];
    leftmost = new int[h];
    rightmost = new int[h];
    sum = 0;
    for (int i = 0; i < w*h; i++) {
        unsigned int c;
        fscanf(pgmfile, " %d", &c);
        if (!feof(pgmfile)) {
            int x = i%w, y = h - i / w - 1;
            bmp[y * w + x] = (1 + taint) * grades - c;
            sum += bmp[y * w + x];
        } else {
            fprintf(stderr,
                    "Sorry, failed to read all image information\n");
            exit(1);
        }
    }
    wxh_inv = 1.0 / (w * h);
    avg = sum * wxh_inv;
    for (y = 0; y < h; y++) {
        accum[y * w1] = 0;
        for (x = 0; x < w; x++) {
            bmp[y*w + x] /= avg;                                                // Make average = 1, hence sum = w * h.
            accum[y * w1 + (x+1)] = accum[y * w1 + x] + bmp[y*w + x];
        }
        accum[y * w1 + w + 1] = accum[y * w1 + w];
    }
}

/*
 * Using CCW sense,
 * hence outside (uncovered) is right hand when going from p1 to p2.
 * To simplify the code, p2 is included in excluded set, then removed later.
 * For left/right markings only intermediate pixels are considered,
 * corners should be processed later by calling routine.
 * The logic is simple and fixed per quarter, not octant.
 * leftmost/rightmost is set immediately before or after y is changed.
 */
double CDensityMap::digitizeEdge(Point p1, Point p2) {
    if (samePixel(p1, p2)) return 0;
    double xx = p2.x - p1.x, yy = p2.y - p1.y;
    double uncovered, excess;
    double logicalLength = 0;                                                   // Length of edge weighted (piecewise) by pixel densities
    int x = int(p1.x), y = int(p1.y);
    int x2 = int(p2.x), y2 = int(p2.y);
    sentinelx = x2; sentinely = y2;
    double dx, dy;
    if (yy == 0) {                                                              // Horizontal
        if (xx > 0) {                                                           // To the right
            dy = frac(p1.y);
            uncovered = dy;
            logicalLength += (1 - frac(p1.x)) * getpixel(x, y);
            for (x++; x < x2; x++) {
                exclude(x, y, uncovered);
                logicalLength += getpixel(x, y);
            }
            logicalLength += frac(p2.x) * getpixel(x, y);
        }
        else {                                                                  // To the left
            dy = 1 - frac(p1.y);
            uncovered = dy;
            logicalLength += frac(p1.x) * getpixel(x, y);
            for (x--; x >= x2; x--) {
                exclude(x, y, uncovered);
                logicalLength += getpixel(x, y);
            }
            logicalLength += (1 - frac(p2.x)) * getpixel(x, y);
        }
    }
    else if (yy > 0) {
        if (xx > 0) {                                                           // Quarter 1: x++, y++; rows left at rightmost
            if (xx > yy) {                                                      // Octant 1: x++, y+
                double m = yy / xx;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dy = frac(p1.y) + m * (1 - frac(p1.x));
                if (dy >= 1) {
                    dy -= 1;
                    y++;
                    uncovered = dy * dy * half_inv_m;
                    exclude(x, y, uncovered);
                }
                while (x < x2) {
                    x++;
                    dy += m;                                                    // Now dy is at right edge
                    uncovered = dy - half_m;                                    // Subtract triangle
                    if (dy >= 1) {                                              // We exited upwards
                        dy -= 1;                                                // Correct dy
                        excess = dy * dy * half_inv_m;                          // Area of excess triangle on top
                        uncovered -= excess;                                    // Subtract it for current pixel
                        exclude(x, y, uncovered);
                        rightmost[y] = x;
                        y++;                                                    // Step to pixel above
                        if (y > y2) break;
                        uncovered = excess;                                     // The excess triangle is mow the uncovered part
                    }
                    exclude(x, y, uncovered);
                }
            }
            else {                                                              // Octant 2: x+, y++
                double m = xx / yy;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dx = (1 - frac(p1.x)) - m * (1 - frac(p1.y));
                if (dx < 0) {
                    uncovered = 1 - dx * dx * half_inv_m;
                    x++;
                    exclude(x, y, uncovered);
                    dx += 1;
                }
                while (y < y2) {
                    rightmost[y] = x;
                    y++;
                    uncovered = dx - half_m;
                    if (dx < m) {
                        excess = dx * dx * half_inv_m;
                        exclude(x, y, excess);
                        x++;
                        if (x > x2) break;
                        dx += 1;
                        uncovered += 1 - excess;
                    }
                    exclude(x, y, uncovered);
                    dx -= m;
                }
            }
        }
        else if (xx == 0) {                                                     // Vertical upwards; rows entered/left at rightmost
            dx = 1 - frac(p1.x);
            uncovered = dx;
            for (y++; y <= y2; y++) {
                exclude(x, y, uncovered);
                rightmost[y] = x;
            }
        }
        else {                                                                  // Quarter 2: x--, y++; rows entered at rightmost
            xx = -xx;
            if (xx < yy) {                                                      // Octant 3: x-, y++
                double m = xx / yy;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dx = (1 - frac(p1.x)) + m * (1 - frac(p1.y));
                if (dx >= 1) {
                    dx -= 1;
                    x--;                                                        // rightmost should have been set
                    uncovered = dx * dx * half_inv_m;
                    exclude(x, y, uncovered);
                }
                while (y < y2) {
                    y++;
                    rightmost[y] = x;
                    dx += m;                                                    // Now dx is at top edge
                    uncovered = dx - half_m;                                    // Subtract triangle
                    if (dx >= 1) {                                              // We exited on the left
                        dx -= 1;                                                // Correct dx
                        excess = dx * dx * half_inv_m;                          // Area of excess triangle on the left
                        uncovered -= excess;                                    // Subtract it for current pixel
                        exclude(x, y, uncovered);
                        x--;                                                    // Step to pixel on the left
                        if (x < x2) break;
                        uncovered = excess;                                     // The excess triangle is mow the uncovered part
                    }
                    exclude(x, y, uncovered);
                }
            }
            else {                                                              // Octant 4: x--, y+
                double m = yy / xx;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dy = (1 - frac(p1.y)) - m * frac(p1.x);
                if (dy < 0) {
                    uncovered = 1 - dy * dy * half_inv_m;
                    y++;
                    rightmost[y] = x;
                    exclude(x, y, uncovered);
                    dy += 1;
                }
                while (x > x2) {
                    x--;
                    uncovered = dy - half_m;
                    if (dy < m) {
                        excess = dy * dy * half_inv_m;
                        exclude(x, y, excess);
                        y++;
                        if (y > y2) break;
                        rightmost[y] = x;
                        dy += 1;
                        uncovered += 1 - excess;
                    }
                    exclude(x, y, uncovered);
                    dy -= m;
                }
            }
        }
    }
    else {
        yy = -yy;
        if (xx > 0) {                                                           // Quarter 4: x++, y--; rows entered at leftmost
            if (xx > yy) {                                                      // Octant 8: x++, y-
                double m = yy / xx;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dy = frac(p1.y) - m * (1 - frac(p1.x));
                if (dy < 0) {
                    uncovered = 1 - dy * dy * half_inv_m;
                    y--;
                    leftmost[y] = x;
                    exclude(x, y, uncovered);
                    dy += 1;
                }
                while (x < x2) {
                    x++;
                    uncovered = dy - half_m;
                    if (dy < m) {
                        excess = dy * dy * half_inv_m;
                        exclude(x, y, excess);
                        y--;
                        if (y < y2) break;
                        leftmost[y] = x;
                        dy += 1;
                        uncovered += 1 - excess;
                    }
                    exclude(x, y, uncovered);
                    dy -= m;
                }
            }
            else {                                                              // Octant 7
                double m = xx / yy;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dx = frac(p1.x) + m * frac(p1.y);
                if (dx >= 1) {
                    dx -= 1;
                    x++;
                    uncovered = dx * dx * half_inv_m;
                    exclude(x, y, uncovered);
                }
                while (y > y2) {
                    y--;
                    leftmost[y] = x;
                    dx += m;                                                    // Now dx is at bottom edge
                    uncovered = dx - half_m;                                    // Subtract triangle
                    if (dx >= 1) {                                              // We exited on the right
                        dx -= 1;                                                // Correct dx
                        excess = dx * dx * half_inv_m;                          // Area of excess triangle on the right
                        uncovered -= excess;                                    // Subtract it for current pixel
                        exclude(x, y, uncovered);
                        x++;                                                    // Step to pixel on the right
                        if (x > x2) break;
                        uncovered = excess;                                     // The excess triangle is mow the uncovered part
                    }
                    exclude(x, y, uncovered);
                }
            }
        }
        else if (xx == 0) {                                                     // Vertical downwards; rows entered/left at leftmost
            x = int(p1.x);
            dx = frac(p1.x);
            uncovered = dx;
            for (y--; y >= y2; y--) {
                exclude(x, y, uncovered);
                leftmost[y] = x;
            }
        }
        else {                                                                  // Quarter 3: x--, y--; rows left at leftmost
            xx = fabs(xx);
            if (xx < yy) {                                                      // Octant 6
                double m = xx / yy;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dx = frac(p1.x) - m * frac(p1.y);
                if (dx < 0) {
                    uncovered = 1 - dx * dx * half_inv_m;
                    x--;
                    exclude(x, y, uncovered);
                    dx += 1;
                }
                while (y > y2) {
                    leftmost[y] = x;
                    y--;
                    uncovered = dx - half_m;
                    if (dx < m) {
                        excess = dx * dx * half_inv_m;
                        exclude(x, y, excess);
                        x--;
                        if (x < x2) break;
                        dx += 1;
                        uncovered += 1 - excess;
                    }
                    exclude(x, y, uncovered);
                    dx -= m;
                }
            }
            else {                                                              // Octant 5
                double m = yy / xx;
                double half_m(0.5*m), half_inv_m(0.5/m), sec(sqrt(1+m*m));
                dy = (1 - frac(p1.y)) + m * frac(p1.x);
                if (dy >= 1) {
                    dy -= 1;
                    y--;
                    uncovered = dy * dy * half_inv_m;
                    exclude(x, y, uncovered);
                }
                while (x > x2) {
                    x--;
                    dy += m;                                                    // Now dy is at right edge
                    uncovered = dy - m / 2;                                     // Subtract triangle
                    if (dy >= 1) {                                              // We exited upwards
                        dy -= 1;                                                // Correct dy
                        excess = dy * dy * half_inv_m;                          // Area of excess triangle on top
                        uncovered -= excess;                                    // Subtract it for current pixel
                        exclude(x, y, uncovered);
                        leftmost[y] = x;
                        y--;                                                    // Step to pixel above
                        if (y < y2) break;
                        uncovered = excess;                                     // The excess triangle is mow the uncovered part
                    }
                    exclude(x, y, uncovered);
                }
            }
        }
    }
    return 0;
}

double CDensityMap::mass(Points p) {
    int n = p.size();
    for (int i = 0; i < n; i++) {
        p[i].x *= w;
        p[i].y *= h;
    }
    bool singlePixel = true;
    for (int i = 0; singlePixel && (i < n); i++) {
        if (!samePixel(p[i], p[(i+n-1)%n]))
            singlePixel = false;
    }
    if (singlePixel) return area(p) * getpixel(int(p[0].x), int(p[0].y));
    int yb = h, yt = 0;                                                         // Search for a bounding box.
    double mass = 0;
    for (int i = 0; i < n; i++) {
        yb = std::min(yb, int(p[i].y));
        yt = std::max(yt, int(p[i].y));
    }
    for (int y = yb; y <= yt; y++) {
        leftmost[y] = w;
        rightmost[y] = 0;
    }
    excluded = 0;                                                               // Reset excluded mass accumulator
    for (int i = 0; i < n; i++) {
        digitizeEdge(p[i], p[(i+1)%n]);                                         // mark lefts/rights and accumulate excluded mass
    }
    for (int i = 0; i < n; i++) {
        int x = int(p[i].x), y = int(p[i].y);
        leftmost[y] = std::min(leftmost[y], x);
        rightmost[y] = std::max(rightmost[y], x);
        if (!samePixel(p[i], p[(i+n-1)%n])) {
            double vertexCoverage = vertexPixelArea(p, i);
            excluded += (1 - vertexCoverage) * getpixel(x, y);
        }
    }
    for (int y = yb; y <= yt; y++) {
        mass += accum[ y*w1 + rightmost[y] + 1] - accum[ y*w1 + leftmost[y] ];
    }
    return (mass - excluded) * wxh_inv;
}

double CDensityMap::mass(const Point &p1, const Point &p2, const Point &p3) {
    Points p = {p1, p2, p3};
    if (crossProduct(p[0], p[1], p[2]) < 0) std::swap(p[1], p[2]);
    return mass(p);
}
