/* Produce pinwheel curve designed by Henry Segerman.
 * 2015-10-16: Created by Abdalla Ahmed.
 * 2020-06-20: Revised by Abdalla Ahmed for CGVC submission.
 */

#include "CDensityMap.h"
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <cmath>
#include <time.h>
#include <getopt.h>
#include <string>
#include <cairo/cairo.h>

Point center = {2./3., 1./3.};
Points vertices;
CDensityMap *map;
int maxDepth;
double threshold, lineWeight, lineWidth;
int iterations = 16;
bool smooth = false;
bool stepWise = false;                                                          // Divide the line inside the triangle in (true) three steps or (false) at once
bool showTriangles = false;
bool pointsOnly = false;
bool quantize = false;
inline Point& currentPoint() { return vertices.back(); }


struct Vector {
    double x, y;
    double length() { return std::sqrt(x*x + y*y); }
};

Vector operator-(Point p1, Point p2) { return {p1.x - p2.x, p1.y - p2.y}; }

Vector operator*(double s, Vector v) { return {s * v.x, s * v.y}; }

Point operator+(Point p, Vector v) { return {p.x + v.x, p.y + v.y}; }

double totalLength(Points p) {                                                  // Length of a sequence of line segments.
    double sum(0);
    for (int i = 0; i < p.size() - 1; i++) {
        sum += (p[i + 1] - p[i]).length();
    }
    return sum;
}

inline Point average(const Point &p1, const Point &p2, double t) {              // Weighted average of two points: (1 - t).p1 + t.p2
    return p1 + t * (p2 - p1);
}

inline Points average(const Points &p1, const Points &p2, double t) {           // Weighted average of two sequences of lines segments: (1 - t).p1 + t.p2
    Points p(p1.size());
    for (int i = 0; i < p1.size(); i++) {
        p[i] = average(p1[i], p2[i], t);
    }
    return p;
}

inline Points trapezoid(
    const Points &min, const Points &max, double optimalLength
) {
    double tLess(0), tMore(1), t;
    Points p;
    double length;
    for (int i = 0; i < iterations; i++) {                                      // Use trapezoid approximation between min and max
        t = 0.5 * (tLess + tMore);
        p = average(min, max, t);
        length = totalLength(p) + (p[0] - currentPoint()).length();
        if (length > optimalLength) tMore = t;
        else tLess = t;
    }
    return p;
}

struct TPSMatrix {                                                              // Similar to Postscript matrix.
    double a, b, c, d, tx, ty;
    TPSMatrix concat(TPSMatrix transformation) {
        TPSMatrix result;
        result.a = transformation.a * a + transformation.b * c;
        result.b = transformation.a * b + transformation.b * d;
        result.c = transformation.c * a + transformation.d * c;
        result.d = transformation.c * b + transformation.d * d;
        result.tx = transformation.tx * a + transformation.ty * c + tx;
        result.ty = transformation.tx * b + transformation.ty * d + ty;
        return result;
    };
    double det() {return a * d - b * c;};
};

TPSMatrix M[] = {                                                               // Refer to squared notebook.
    {-0.2, -0.4, -0.4,  0.2, 0.4, 0.8},                                         // 4
    { 0.4, -0.2, -0.2, -0.4, 0.2, 0.4},                                         // 1
    { 0.4, -0.2,  0.2,  0.4, 0.2, 0.4},                                         // 0
    {-0.4,  0.2, -0.2, -0.4, 1.2, 0.4},                                         // 2
    { 0.4, -0.2, -0.2, -0.4, 1.2, 0.4},                                         // 3
};

inline Point transform(const Point &p, const TPSMatrix &m) {                    // Please note that b and c are swapped in PS convention.
    double X = m.a * p.x + m.c * p.y + m.tx;
    double Y = m.b * p.x + m.d * p.y + m.ty;
    return {X, Y};
};

inline Point transform (double x, double y, const TPSMatrix &m) {
    Point p = {x, y};
    return transform(p, m);
}

Points transform(Points p, const TPSMatrix m) {
    for (int i = 0; i < p.size(); i++) {
        p[i] = transform(p[i], m);
    }
    return p;
}

Point itransform(Point p, TPSMatrix m) {
    double x = p.x - m.tx, y = p.y - m.ty;
    double X = (m.d * x - m.c * y) / m.det();
    double Y = (-m.b * x + m.a * y) / m.det();
    return {X, Y};
}

struct TPinwheelTile {
    TPSMatrix matrix;
    bool BtoC;                                                                  // Order of children goes from (true) B to C or (false) C to B?
    Point A() { return transform(0, 0, matrix); }
    Point C() { return transform(2, 0, matrix); }
    Point B() { return transform(0, 1, matrix); }
    Point mid() { return transform(center, matrix); }
    double b() { return (A() - C()).length(); }
    double a() { return (C() - B()).length(); }
    double c() { return (B() - A()).length(); }
    bool drawLineQuantized(int depth);
    bool drawLineAtOnce(int depth);
    bool drawLineStepwise(int depth);
    bool drawLine(int depth);
    double enclosedMass() {
        return map->mass(A(), C(), B());
    }
};

static const Points maxBtoC = {                                                 // Centroids of children ordered from B to C
    { 2./15., 9./15.},
    { 6./15., 2./15.},
    { 8./15., 6./15.},
    {13./15., 6./15.},
    {21./15., 2./15.}
};

static const Points maxCtoB = {                                                 // Centroids of children ordered from B to C
    {21./15., 2./15.},
    {13./15., 6./15.},
    { 8./15., 6./15.},
    { 6./15., 2./15.},
    { 2./15., 9./15.}
};

bool TPinwheelTile::drawLine(int depth) {
    if (quantize) return drawLineQuantized(depth);
    if (stepWise) return drawLineStepwise(depth);
    return drawLineAtOnce(depth);
}

bool TPinwheelTile::drawLineQuantized(int depth) {
    double optimalLength = enclosedMass() * threshold;
    Point center = mid();
    double minLength = (center - currentPoint()).length();
    if (minLength > optimalLength || depth >= maxDepth) {                   // One segment is more than needed.
        vertices.push_back(center);
        return true;
    }
    else return false;
}

bool TPinwheelTile::drawLineAtOnce(int depth) {
    double optimalLength = enclosedMass() * threshold;
    Point center = mid();
    double minLength = (center - currentPoint()).length();
    if (quantize) {
        if (minLength > optimalLength || depth >= maxDepth) {                   // One segment is more than needed.
            vertices.push_back(center);
            return true;
        }
        else return false;
    }
    if (minLength > optimalLength) {                                         // One segment is more than needed.
        vertices.push_back(center);
        return true;
    }
    Points max = BtoC ? transform(maxBtoC, matrix) :
                             transform(maxCtoB, matrix);
    double maxLength = totalLength(max) + (max[0] - currentPoint()).length();
    if (maxLength > optimalLength) {                                            // We can match the mass
        Points min(5);                                                          // Split the line segment from currentpoint to centerpoint into five points
        for (int i = 0; i < 5; i++) {
            min[i] = average(currentPoint(), center, 0.2 * i);
        }
        Points p = trapezoid(min, max, optimalLength);
        vertices.insert(vertices.end(), p.begin(), p.end());
        return true;
    }
    if (depth >= maxDepth) {                                                    // We have to live with our maximum.
        for (int i = 0; i < 5; i++) {
            vertices.push_back(max[i]);
        }
        return true;
    }
    return false;
}

bool TPinwheelTile::drawLineStepwise(int depth) {
    double optimalLength = enclosedMass() * threshold;
    if ((mid() - currentPoint()).length() > optimalLength) {                   // One segment is more than needed.
        vertices.push_back(mid());
        return true;
    }
    Points centers =
        BtoC ? transform(maxBtoC, matrix) : transform(maxCtoB, matrix);
    double length;
    Points min, max;
    if (
        (
            (mid() -   centers[0]).length() +
            (centers[0] - currentPoint()).length()
        ) > optimalLength
    ) {                                                                         // Two segments are sufficient
        min = {average(mid(), currentPoint(), 0.5), mid()};
        max = {centers[0], mid()};
    }
    else if (
        (
            (centers[4] -   centers[0]).length() +
            (centers[0] - currentPoint()).length()
        ) > optimalLength
    ) {                                                                         // Two segments are sufficient
        min = {centers[0], mid()};
        max = {centers[0], centers[4]};
    }
    else if (
        (
            (centers[4] -   centers[2]).length() +
            (centers[2] -   centers[0]).length() +
            (centers[0] - currentPoint()).length()
        ) > optimalLength
    ) {                                                                         // three segments are sufficient
        min = {centers[0], average(centers[0], centers[4], 0.5), centers[4]};
        max = {centers[0], centers[2], centers[4]};
    }
    else if (
        (
            totalLength(centers) +
            (centers[0] - currentPoint()).length()
        ) > optimalLength
    ) {                                                                         // All the line segments are sufficient
        min = {
            centers[0],
            average(centers[0], centers[2], 0.5),
            centers[2],
            average(centers[2], centers[4], 0.5),
            centers[4]
        };
        max = centers;
    }
    else if (depth >= maxDepth) {                                               // Insufficient, but maximum depth reached: we have to live with out max
        vertices.insert(vertices.end(), centers.begin(), centers.end());
        return true;                                                            // To stop further subdivision.
    }
    else {
        return false;
    }
    Points p = trapezoid(min, max, optimalLength);
    vertices.insert(vertices.end(), p.begin(), p.end());
    return true;
}

std::vector<TPinwheelTile> triangles;

void subdivide(TPinwheelTile topTile) {
    struct TRecord {
        TPinwheelTile tile;
        int depth;
    };
    std::vector<TRecord> stack;
    stack.reserve(pow(5, maxDepth));
    stack.push_back({topTile, 0});
    do {
        TPinwheelTile t = stack.back().tile;
        int depth = stack.back().depth;
        stack.pop_back();
        if (!t.drawLine(depth)) {                                               // Push 5 children in stack
            TPinwheelTile t0 = {t.matrix.concat(M[0]),  t.BtoC};                // Child inheriting vertex B: inherits the same ordering for its children
            TPinwheelTile t1 = {t.matrix.concat(M[1]),  t.BtoC};                // Child sharing vertex A with t0: same ordering
            TPinwheelTile t2 = {t.matrix.concat(M[2]), !t.BtoC};                // Child not sharing any edge with parent: inverted ordering
            TPinwheelTile t3 = {t.matrix.concat(M[3]), !t.BtoC};                // Child sharing side a but no vertices: inverted ordering
            TPinwheelTile t4 = {t.matrix.concat(M[4]),  t.BtoC};                // Child sharing vertex C: same ordering as the parent
            depth++;                                                            // We are getting deeper in recursion
            if (t.BtoC) {                                                       // Push children in stack in reverse (C to B) order so that the one at B is processed first
                stack.push_back({t4, depth});
                stack.push_back({t3, depth});
                stack.push_back({t2, depth});
                stack.push_back({t1, depth});
                stack.push_back({t0, depth});
            }
            else {                                                              // Push children in stack in a B to C order so that they are processed in the desired C to B order
                stack.push_back({t0, depth});
                stack.push_back({t1, depth});
                stack.push_back({t2, depth});
                stack.push_back({t3, depth});
                stack.push_back({t4, depth});
            }
        }
        else if (showTriangles) {
            triangles.push_back(t);
        }
    } while(!stack.empty());
}

void generateVertices() {
    vertices.clear();
    vertices.reserve(20 * pow(5, maxDepth));                                    // 4 grand parents x 5 ^ maxDepth children x 5 vertices.
    vertices.push_back({0.5, 1});
    clock_t t0 = clock();
    subdivide({{0, -0.5, 0.5, 0, 0.0, 1.0}, true});
    subdivide({{0,  0.5,-0.5, 0, 0.5, 0.0}, true});
    subdivide({{0,  0.5, 0.5, 0, 0.5, 0.0}, false});
    subdivide({{0, -0.5,-0.5, 0, 1.0, 1.0}, false});
    clock_t t1 = clock();
    double totalTime = (double)(t1 - t0) / CLOCKS_PER_SEC;
    int n = vertices.size();
    fprintf(
        stderr, "Generated %d vertices in %.6fs (%'d vertex per second)..\n",
            (n - 1), totalTime, (int)((n - 1)/totalTime)
    );
}

void plotEPS (const char *fileName) {
    FILE *file = fopen(fileName, "w");
    fprintf(
        file,
        "%%!PS-Adobe-3.0 EPSF-3.0\n"
        "%%%%BoundingBox: 0 0 1024 1024\n"
        "/lw %f def\n"
        "/smooth %s def\n"
        "/margin lw def\n"
        "1024 1 margin 2 mul add div dup scale\n"
        "margin dup translate\n"
        "lw setlinewidth\n"
        "1 setlinecap\n"
        "1 setlinejoin\n"
        "/t {moveto lineto lineto closepath stroke} def\n"
        "%f %f moveto\n"
        "/mean {\n"
        "    3 -1 roll add 0.5 mul 3 1 roll add 0.5 mul exch\n"
        "} bind def\n"
        "smooth {\n"
        "    currentpoint 2 copy %f %f mean moveto\n"
        "    currentpoint 4 2 roll\n"
        "    /l {\n"
        "        2 copy 6 4 roll\n"
        "        2 copy 2 copy\n"
        "        8 -2 roll\n"
        "        mean\n"
        "        curveto\n"
        "    } def\n"
        "} {\n"
        "    /l {lineto} def\n"
        "} ifelse\n",
        lineWidth,
        smooth ? "true" : "false",
        vertices[0].x, vertices[0].y,
        vertices.back().x, vertices.back().y
    );
    if (showTriangles) {
        fprintf(
            file, "gsave 1 0.8 0.5 setrgbcolor lw 0.5 mul setlinewidth\n"
        );
        for (int i = 0; i < triangles.size(); i++) {
            fprintf(
                file,
                "%0.6f %0.6f %0.6f %0.6f %0.6f %0.6f t\n",
                triangles[i].C().x, triangles[i].C().y,
                triangles[i].B().x, triangles[i].B().y,
                triangles[i].A().x, triangles[i].A().y
            );
        }
        fprintf(file, "grestore\n");
    }
    if (pointsOnly) {
        fprintf(
            file,
            "/p {newpath lw 0 360 arc fill} def\n"
        );
        for (int i = 1; i < vertices.size(); i++) {
            fprintf(file, "%0.6f %0.6f p\n", vertices[i].x, vertices[i].y);
        }
        fprintf(file, "showpage\n");
    }
    else {
        for (int i = 1; i < vertices.size(); i++) {
            fprintf(file, "%0.6f %0.6f l\n", vertices[i].x, vertices[i].y);
        }
        fprintf(
            file,
            "smooth {2 copy 6 -2 roll curveto} if\n"
            "closepath stroke\n"
            "showpage\n"
        );
    }
    fclose(file);
}

void plotPNG(const char *fileName, int imageSize) {
    cairo_surface_t *surface = cairo_image_surface_create(
        CAIRO_FORMAT_RGB24, imageSize, imageSize
    );
    cairo_t *cr = cairo_create(surface);

    double scale = imageSize / (1 + 2 * lineWidth);
    cairo_scale (cr, scale, -scale);
    cairo_translate(cr, lineWidth, -1 - lineWidth);
    cairo_set_line_width (cr, lineWidth);
    cairo_set_line_cap(cr, CAIRO_LINE_CAP_ROUND);
    cairo_set_line_join(cr, CAIRO_LINE_JOIN_ROUND);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    if (smooth) {
        Point p0 = average(vertices[0], vertices.back(), 0.5);
        cairo_move_to(cr, p0.x, p0.y);
        vertices.push_back(vertices[0]);                                        // Make a full round
        for (int i = 0; i < vertices.size() - 1; i++) {
            Point p = average(vertices[i], vertices[i+1], 0.5);
            cairo_curve_to(
                cr,
                vertices[i].x, vertices[i].y,
                vertices[i].x, vertices[i].y,
                p.x, p.y
            );
        }
        vertices.pop_back();                                                    // Remove the vertex added earlier for a full round
    }
    else {
        cairo_move_to(cr, vertices[0].x, vertices[0].y);
        for (int i = 1; i < vertices.size(); i++) {
            cairo_line_to(cr, vertices[i].x, vertices[i].y);
        }
    }
    cairo_close_path(cr);
    cairo_stroke (cr);

    cairo_destroy (cr);
    cairo_surface_write_to_png (surface, fileName);
    cairo_surface_destroy (surface);

}

const char *USAGE_MESSAGE =
    "Usage: %s <pgmfile> <threshold> <output file name> [options]\n"
    "Options:\n"
    "    -s                 Make a smooth line\n"
    "    -z <taint>         Add a fractional offset to all pixels to obtain no zero pixels\n"
    "    -w <image width>   Width of png image. Default is 1024\n"
    "    -t                 Draw the triangles in the rendering\n"
    "    -q                 Qantize to full octaves only\n"
    "    -p                 Only show the points\n"
    "    -P                 Stepwise subdivision\n";

int main(int argc,char **argv) {
    srand(time(NULL));
    int opt;                                                                    // For use by getopt, the command line options utility
    double taint = 0;
    int imageSize = 1024;
    while ((opt = getopt(argc, argv, "sz:w:Ptpq")) != -1) {
        switch (opt) {
            case 's': smooth = true; break;
            case 'z': taint = atof(optarg); break;
            case 'w': imageSize = atoi(optarg); break;
            case 'P': stepWise = true; break;
            case 't': showTriangles = true; break;
            case 'p': pointsOnly = true; break;
            case 'q': quantize = true; break;
            default: fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
        }
    }
    if (optind > argc - 3) {
        fprintf(stderr, USAGE_MESSAGE, argv[0]);
        exit(1);
    }
    char *inputFileName = argv[optind];
    threshold = atof(argv[optind + 1]);
    std::string outputFileName = argv[optind + 2];
    lineWeight = 1.0 / threshold;
    maxDepth = round(log(threshold * threshold) / log(5));
    lineWidth = (1./15) * lineWeight;                                           // Half of the nearest distance the curve comes to itself.
    fprintf(stderr, "maxDepth = %d, threshold = %f\n", maxDepth, threshold);
    map = new CDensityMap(inputFileName, taint);
    generateVertices();
    if (
        outputFileName.rfind(".eps") == outputFileName.length() - 4 ||
        outputFileName.rfind(".EPS") == outputFileName.length() - 4
    ) {
        fprintf(
            stderr, "Writing halftoned image to %s\n", outputFileName.c_str()
        );
        plotEPS(outputFileName.c_str());
    }
    else if (
        outputFileName.rfind(".png") == outputFileName.length() - 4 ||
        outputFileName.rfind(".PNG") == outputFileName.length() - 4
    ) {
        fprintf(
            stderr, "Writing halftoned image to %s\n", outputFileName.c_str()
        );
        plotPNG(outputFileName.c_str(), imageSize);
    }
    return 0;
}
